from asyncio import Event as _Event, sleep as _sleep
from collections.abc import Callable as _Callable
from io import BytesIO as _BytesIO, StringIO as _StringIO
from typing import Any as _Any, TypedDict as _TypedDict

from aiohttp import (
    ClientConnectorDNSError as _ClientConnectorDNSError,
    ClientResponseError as _ClientResponseError,
)
from aiohutils.pd import html_to_df as _html_to_df
from numpy import nan as _nan
from pandas import concat as _concat, to_numeric as _to_numeric

from tsetmc import (
    MarketState,
    _csv2df,
    _DataFrame,
    _get_data,
    _get_par_tree,
    _jstrptime,
    _logger,
    _parse_market_state,
    _save_last_content,
)

_BEST_LIMITS_NAMES = ('ins_code', 'number', 'zo', 'zd', 'pd', 'po', 'qd', 'qo')
_COMMON_DTYPES = {
    'heven': 'int32',
    'pf': 'int64',
    'pc': 'int64',
    'pl': 'int64',
    'tno': 'int64',
    'tvol': 'int64',
    'tval': 'int64',
    'pmin': 'int64',
    'pmax': 'int64',
}
_PRICE_DTYPES_23 = {
    'ins_code': 'string',
    'isin': 'string',
    'l18': 'string',
    'l30': 'string',
    **_COMMON_DTYPES,
    'py': 'int64',
    'eps': 'float64',
    'bvol': 'int64',
    'visitcount': 'int64',
    # 0-7 http://redirectcdn.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83
    'flow': 'int16',
    # 1-98, see tsetmc.general.cs_codes()
    'cs': 'string',
    'tmax': 'float64',
    'tmin': 'float64',
    'z': 'int64',
    # 67-701 http://redirectcdn.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83
    'yval': 'string',
}
_PRICE_DTYPES_25 = _PRICE_DTYPES_23 | {'predtran': 'float64', 'buyop': 'Int64'}
_PRICE_UPDATE_COLUMNS = {'ins_code': 'string', **_COMMON_DTYPES}


class MarketWatchInit(_TypedDict):
    prices: _DataFrame
    best_limits: _DataFrame
    market_state: MarketState
    refid: int


def _unstack_best_limits(bl: _DataFrame) -> _DataFrame:
    # merge multiple rows sharing the same `row` number into one row.
    # a fascinating solution from https://stackoverflow.com/a/53563551/2705757
    bl = bl.unstack()  # type: ignore
    bl.columns = [f'{name}{number}' for name, number in bl.columns]
    return bl


async def market_watch_init(
    *, market_state=True, prices=True, best_limits=True, join=True
) -> MarketWatchInit:
    """Return the market status which are the info used in creating filters.

    If `join` is True, `best_limits` will be joined to `prices`, otherwise
        it will be returned as a separate dataframes.

    For more information about filters see:
        http://tsetmc.com/Loader.aspx?ParTree=15131F
    For the meaning of column names in the returned DataFrame refer to:
        http://tsetmc.com/Site.aspx?ParTree=151713
        For persian translation of `flow` and `yval` codes use
        `tsetmc.docs.instrument`.
        `heven` is the time of the last transaction in HHMMSS format.
    """
    text = await _get_data('MarketWatchInit.aspx?h=0&r=0', fa=True)
    try:
        _, market_state_str, states, price_rows, refid = text.split('@')
    except ValueError:
        # ValueError: not enough values to unpack (expected 5, got 1)
        _logger.error(text)
        raise

    result: dict = {'refid': int(refid)}
    if prices:
        result['prices'] = price_df = _csv2df(
            _StringIO(states),
            names=_PRICE_DTYPES_25,  # type: ignore
            index_col='ins_code',
            dtype=_PRICE_DTYPES_25,
        )
    if best_limits:
        result['best_limits'] = bl = _csv2df(
            _StringIO(price_rows),
            names=_BEST_LIMITS_NAMES,
            dtype={'ins_code': 'string'},
            index_col=('ins_code', 'number'),
        )
    if join and prices and best_limits:
        # noinspection PyUnboundLocalVariable
        bl = _unstack_best_limits(bl)  # type: ignore
        # noinspection PyUnboundLocalVariable
        joined = bl.join(price_df)  # type: ignore
        # joined_df.index = to_numeric(joined_df.index, downcast='unsigned')
        result['prices'] = joined
    if market_state:
        result['market_state'] = _parse_market_state(market_state_str)
    return result  # type: ignore


class MarketWatchPlus(_TypedDict):
    new_prices: _DataFrame
    price_updates: _DataFrame
    best_limits: _DataFrame
    messages: list[str]
    market_state: MarketState
    refid: int


async def market_watch_plus(
    heven: int,
    refid: int,
    *,
    messages=True,
    market_state=True,
    new_prices=True,
    price_updates=True,
    best_limits=True,
    best_limits_prepare_join=True,
) -> MarketWatchPlus:
    # See dev/tsetmc_source_files/market_watch.html
    # for how the response is parsed in the browser.
    text = await _get_data(
        f'MarketWatchPlus.aspx?h={5 * (heven // 5)}&r={25 * (refid // 25)}',
        fa=True,
    )
    try:
        (
            handle_msg,
            update_fast_view,
            inst_price,
            best_limit,
            str_refid,
        ) = text.split('@')
    except ValueError as e:
        _save_last_content(f'{e!r}')
        raise e
    result = {}
    if messages:
        # whenever a new id appears, users should try to fetch new messages
        # using relevant functions
        # todo: implement functions to fetch messages using message ids
        # NewMsgNotification, NewInsStateNotification, NewCodalNotification
        result['messages'] = handle_msg.split(',')
    if market_state:
        if update_fast_view != '':
            result['market_state'] = _parse_market_state(update_fast_view)
    if new_prices or price_updates:
        inst_prices = [ip.split(',') for ip in inst_price.split(';')]
        if new_prices:
            lst = [ip for ip in inst_prices if len(ip) != 10]
            twenty_five_cols = not lst or len(lst[0]) == 25
            try:
                # noinspection PyTypeChecker
                # https://github.com/pandas-dev/pandas/issues/57798
                df = _DataFrame(
                    lst,
                    columns=_PRICE_DTYPES_25
                    if twenty_five_cols is True
                    else _PRICE_DTYPES_23,
                    copy=False,
                )
            except ValueError as e:
                _save_last_content(f'{e}')
                raise e
            df['eps'] = df['eps'].replace('', _nan)
            if twenty_five_cols is True:
                df['predtran'] = df['predtran'].replace('', _nan)
                df['buyop'] = df['buyop'].replace('', _nan)
                df = df.astype(_PRICE_DTYPES_25)
            else:
                df = df.astype(_PRICE_DTYPES_23)
            df.set_index('ins_code', inplace=True)
            result['new_prices'] = df
        if price_updates:
            lst = [ip for ip in inst_prices if len(ip) == 10]
            # noinspection PyTypeChecker
            # https://github.com/pandas-dev/pandas/issues/57798
            df = _DataFrame(lst, columns=_PRICE_UPDATE_COLUMNS, copy=False)
            df = df.astype(_PRICE_UPDATE_COLUMNS)
            df.ins_code = df.ins_code.astype('string')
            df.set_index('ins_code', inplace=True)
            result['price_updates'] = df
    if best_limits:
        bl = _csv2df(
            _StringIO(best_limit),
            index_col=('ins_code', 'number'),
            names=_BEST_LIMITS_NAMES,
            dtype={'ins_code': 'string'},
        )
        if best_limits_prepare_join:
            bl = _unstack_best_limits(bl)
        result['best_limits'] = bl
    result['refid'] = int(str_refid)
    return result  # type: ignore


def _split_id_rows(content: bytes, id_row_len: int) -> list:
    data = content.split(b';')
    for i, datum in enumerate(data):
        items = datum.split(b',')
        if len(items) == id_row_len:
            id_ = items[0]
        else:
            # noinspection PyUnboundLocalVariable
            items.insert(0, id_)  # type: ignore
        data[i] = items  # type: ignore
    return data


async def closing_price_all() -> _DataFrame:
    """Return price history dataframe.

    For the meaning of column names refer to
        http://cdn.tsetmc.com/Site.aspx?ParTree=151713
    This method returns the same set of information as `[ih]` variable in
        tsetmc filters. See:
            http://tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3197
    """
    content = await _get_data('ClosingPriceAll.aspx')
    data = _split_id_rows(content, id_row_len=11)
    columns = [
        'ins_code',
        'n',
        'pc',
        'pl',
        'tno',
        'tvol',
        'tval',
        'pmin',
        'pmax',
        'py',
        'pf',
    ]
    df = _DataFrame(
        data,
        columns=columns,
        copy=False,
    )
    df[columns[1:]] = df[columns[1:]].apply(_to_numeric)
    df.ins_code = df.ins_code.astype('string')
    df.set_index(['ins_code', 'n'], inplace=True)
    return df


async def client_type_all() -> _DataFrame:
    """Return client types (natural/legal stats) as a DataFrame.

    In column names `n_` prefix stands for natural and `l_` for legal.
    """
    content = await _get_data('ClientTypeAll.aspx')
    df = _csv2df(
        _BytesIO(content),
        names=(
            'ins_code',
            'n_buy_count',
            'l_buy_count',
            'n_buy_volume',
            'l_buy_volume',
            'n_sell_count',
            'l_sell_count',
            'n_sell_volume',
            'l_sell_volume',
        ),
        index_col='ins_code',
        dtype={'ins_code': 'string'},
    )
    return df


async def key_stats() -> _DataFrame:
    """Return key statistics as a DataFrame.

    For the meaning of column names refer to
        http://www.tsetmc.com/Site.aspx?ParTree=151715&LnkIdn=3199 or
        http://cdn.tsetmc.com/Site.aspx?ParTree=151713
    """
    content = await _get_data('InstValue.aspx?t=a')
    data = _split_id_rows(content, id_row_len=3)
    df = _DataFrame(data, columns=('ins_code', 'n', 'value'), copy=False)
    df.set_index(df.pop('ins_code').astype('string'), inplace=True)
    df = df.apply(_to_numeric)
    df = df.pivot(columns='n', values='value')
    df.columns = [f'is{c}' for c in df.columns]
    return df


async def status_changes(top: int | str) -> _DataFrame:
    text = await _get_par_tree(f'15131L&top={top}')
    df = _html_to_df(text)
    df['date'] = [
        _jstrptime(i, format='%Y/%m/%d %H:%M:%S')
        for i in (df['تاریخ'] + ' ' + df['زمان'])
    ]
    df.drop(columns=['تاریخ', 'زمان'], inplace=True)
    return df


class MarketWatch:
    __slots__ = (
        'df',
        'init_callback',
        'init_kwargs',
        'interval',
        'market_state',
        'plus_callback',
        'plus_kwargs',
        'update_event',
    )

    def __init__(
        self,
        *,
        init_kwargs: dict | None = None,
        plus_kwargs: dict | None = None,
        init_callback: _Callable[[MarketWatchInit], _Any] | None = None,
        plus_callback: _Callable[[MarketWatchPlus], _Any] | None = None,
        interval=1,
    ):
        """Create an object that helps with watching the market watch.

        :param init_kwargs: kwargs to be passed to market_watch_init
        :param plus_kwargs: kwargs to be passed to market_watch_plus
        :param init_callback: function to be called with the result of
            market_watch_init.
        :param plus_callback: function to be called with the result of
            market_watch_plus.
        :param interval: The sleep interval.

        Each time a callback returns, self.event will be set. Users can wait
        for this event to get notified of new updates.

        To stop the watch, cancel the task that is running `self.start()`.

        If init_callback is None, then self._default_init_callback and
        self._default_plus_callback will be used which will create
        self.df and keep it up-to-date while the watch is running.
        This is convenient, but note that you may be able to be implement a
        more efficient algorith to gather specific updates by using
        custom callback functions.
        """
        self.interval = interval
        self.init_kwargs: dict = {} if init_kwargs is None else init_kwargs
        self.plus_kwargs: dict = {} if plus_kwargs is None else plus_kwargs
        self.update_event = _Event()
        self.init_callback = (
            self._default_init_callback
            if init_callback is None
            else init_callback
        )
        self.plus_callback = (
            self._default_plus_callback
            if plus_callback is None
            else plus_callback
        )

    def _default_init_callback(self, d: MarketWatchInit):
        self.df = d.get('prices')
        self.market_state = d.get('market_state')

    def _default_plus_callback(self, d: MarketWatchPlus):
        bl = d.get('best_limits')
        if bl is not None:
            self.df.update(bl)

        np = d.get('new_prices')
        if np is not None:
            self.df = _concat([self.df, np])

        pu = d.get('price_updates')
        if pu is not None:
            self.df.update(pu)

        self.market_state = d.get('market_state')

    async def start(self):
        update_event = self.update_event
        set_event = update_event.set
        clear_event = update_event.clear

        while True:
            try:
                mwi = await market_watch_init(**self.init_kwargs)
            except Exception as e:
                _logger.error(f'{e!r} while awaiting market_watch_init')
                await _sleep(self.interval)
                continue
            break

        self.init_callback(mwi)  # noqa: F823
        heven = mwi['prices'].heven.max()
        refid = mwi['refid']
        set_event()
        clear_event()

        while True:
            await _sleep(self.interval)
            try:
                mwp = await market_watch_plus(
                    refid=refid, heven=heven, **self.plus_kwargs
                )
            except (_ClientConnectorDNSError, _ClientResponseError) as e:
                _logger.warning(f'{e!r} while awaiting market_watch_plus')
                continue
            except Exception as e:
                _logger.exception(f'{e!r} while awaiting market_watch_plus')
                continue  # _sleep and retry

            self.plus_callback(mwp)

            refid = mwp['refid']
            heven = max(
                mwp['price_updates'].heven.max(),
                mwp['new_prices'].heven.max(),
            )
            set_event()
            clear_event()



from typing import TypedDict as _TypedDict
from warnings import warn as _warn

from aiohutils.pd import html_to_df as _html_to_df
from lxml.html import fromstring as _html
from pandas import (
    NA as _NA,
    Timestamp as _Timestamp,
    concat as _concat,
    json_normalize as _json_normalize,
)

from tsetmc import (
    Flow as _Flow,
    FlowType as _FlowType,
    _api,
    _DataFrame,
    _get_par_tree,
    _mem_par_tree,
    _numerize,
)


async def boards() -> dict[int, str]:
    """See http://members.tsetmc.com/Loader.aspx?ParTree=121C1913."""
    content = await _mem_par_tree('111C1913')
    iloc = _html_to_df(content, header=0).iloc
    return dict(zip(iloc[:, 0], iloc[:, 1]))


async def cs_codes() -> dict[str, str]:
    """https://members.tsetmc.com/Loader.aspx?ParTree=111C1213"""
    content = await _mem_par_tree('111C1213')
    iloc = _html_to_df(content, header=0).iloc
    return dict(zip(iloc[:, 0], iloc[:, 1]))


async def sectors_summary() -> _DataFrame:
    j = await _api('MarketData/GetSectorsSummary', fa=True)
    return _DataFrame(j['sectorSummeries'])


async def industrial_groups_overview() -> _DataFrame:
    """Return a dataframe of industrial groups.

    The result contains info about each group's price change.
    See: http://members.tsetmc.com/Loader.aspx?ParTree=111C1214
    """
    _warn(
        'industrial_groups_overview is deprecated; use sectors_summary instead.',
        DeprecationWarning,
        2,
    )
    content = await _mem_par_tree('111C1214')
    df = _html_to_df(content)
    show = df[1]
    df.drop(columns=1, inplace=True)
    percents = show.str.extract(
        r"showBar\('[^\[]*',(\d+),(\d+),(\d+),(\d+)\)"
    ).astype('int64')
    df = _concat((df, percents), copy=False, axis=1)
    df.columns = ('group', ':-2', '-2:0', '0:2', '2:')
    return df


async def market_map_data(
    *, market=0, size=9999, sector=0, typeSelected=1, heven=0
) -> _DataFrame:
    j = await _api(
        f'ClosingPrice/GetMarketMap'
        f'?market={market}&size={size}&sector={sector}'
        f'&typeSelected={typeSelected}&hEven={heven}',
        fa=True,
    )
    df = _DataFrame(j)
    return df


async def major_holders_activity() -> _DataFrame:
    text = await _get_par_tree('15131I')
    html = _html(text)
    trs = html.xpath('//tr')

    rows = []
    append_row = rows.append
    for tr in trs[1:]:
        tds = tr.xpath('.//td')
        td0 = tds[0]

        inst_div = td0.xpath('.//div[1]')
        if inst_div:
            inst_div = inst_div[0]
            href = inst_div.xpath('.//a[1]/@href')[0]
            ins_code = href[href.rfind('=') + 1 :]
            l30 = inst_div.text_content()

        holder = td0.xpath('.//li[1]/text()')[0]
        # ins_code and l30 are expected to be bound at this point
        append_row([ins_code, l30, holder, *_parse_tds(tds)])  # type: ignore
    return _DataFrame(
        rows,
        copy=False,
        columns=(
            'ins_code',
            'l30',
            'holder',
            *(
                # the first header is 'شرکت - سهامدار'
                th.text
                for th in trs[0].xpath('.//th')[1:]
            ),
        ),
    )


async def top_industry_groups() -> _DataFrame:
    """http://old.tsetmc.com/Loader.aspx?Partree=15131O"""
    text = await _get_par_tree('15131O')
    df = _html_to_df(text)
    df.columns = ['group', 'mv', 'tno', 'tvol', 'tval']
    _numerize(df, ('mv', 'tvol', 'tval'), 'float64', comma=True)
    return df


def _parse_tds(tds):
    for td in tds[1:]:
        text = td.text
        try:
            yield float(text.replace(',', ''))
        except ValueError:
            if td == '\xa0':
                yield _NA


class MarketOverview(_TypedDict):
    indexChange: float
    indexEqualWeightedChange: float
    indexEqualWeightedLastValue: float
    indexLastValue: float
    lastDataDEven: int
    lastDataHEven: int
    marketActivityDEven: int
    marketActivityHEven: int
    marketActivityQTotCap: float
    marketActivityQTotTran: float
    marketActivityTimestamp: _Timestamp
    marketActivityZTotTran: int
    marketState: str
    marketStateTitle: str
    marketValue: float
    marketValueBase: float


async def market_overview(*, flow: _FlowType = _Flow.BOURSE) -> MarketOverview:
    """tsetmc.com > در یک نگاه"""
    j = await _api(f'MarketData/GetMarketOverview/{flow}')
    overview = j['marketOverview']
    overview['marketActivityTimestamp'] = _Timestamp(
        f'{overview["marketActivityDEven"]}'
        f'{overview["marketActivityHEven"]:>06}'
    )
    return overview


async def related_companies(cs: str) -> dict[str, _DataFrame]:
    j = await _api(f'ClosingPrice/GetRelatedCompany/{cs}')
    j['relatedCompany'] = _json_normalize(j['relatedCompany'])
    j['relatedCompanyThirtyDayHistory'] = _DataFrame(
        j['relatedCompanyThirtyDayHistory'], copy=False
    )
    return j


async def messages(
    *, flow: _FlowType = _Flow.GENERAL, top: int | str = 200
) -> _DataFrame:
    """See also: ``search_messages`` and ``Instrument.messages``."""
    j = await _api(f'Msg/GetMsgByFlow/{flow}/{top}', fa=True)
    df = _DataFrame(j['msg'])
    return df


async def search_messages(*, sh_date: str, term: str) -> _DataFrame:
    """https://tsetmc.com/MsgTop

    :param term: Only return messages containing this term.
    :param sh_date: Solar Hijri date string in 'YYYY-mm-dd' format.

    See also: ``messages`` and ``Instrument.messages``.
    """
    j = await _api(f'Msg/GetMsgByDevenAndLVal18AFC/{sh_date}/{term}', fa=True)
    return _DataFrame(j['msg'])



from ast import literal_eval as _literal_eval
from functools import partial as _partial
from io import BytesIO as _BytesIO, StringIO as _StringIO
from logging import warning as _warning
from pathlib import Path
from re import Match as _Match, findall as _findall, fullmatch as _fullmatch
from typing import Literal as _Literal, overload as _overload
from warnings import warn as _warn

from aiohutils.pd import html_to_df as _html_to_df
from pandas import (
    Timestamp as _Ts,
    read_csv as _read_csv,
    to_datetime as _to_datetime,
)

from tsetmc import (
    _F,
    _FARSI_NORM,
    FlowType as _FlowType,
    InstrumentInfo,
    MarketState,
    _api,
    _csv2df,
    _DataFrame,
    _get,
    _get_data,
    _get_par_tree,
    _jdatetime,
    _jstrptime,
    _numerize,
    _parse_market_state,
    _rc,
    _TypedDict,
)

_j_ymd_parse = _partial(_jstrptime, format='%Y/%m/%d')


_FARSI_NORM_REVERSED = {v: k for k, v in _FARSI_NORM.items()}


# This regex is generated using dev/page_vars_regex_generator.py.
# Fix and update that script before making changes.
_PAGE_VARS = _rc(
    "TopInst='(?P<TopInst>[^,]*)',"
    "LVal18AFC='[^,]*',"
    "DEven='(?P<DEven>[^,]*)',"
    "LSecVal='(?P<LSecVal>[^,]*)',"
    "CgrValCot='(?P<CgrValCot>[^,]*)',"
    "Flow='(?P<Flow>[^,]*)',"
    "InstrumentID='(?P<InstrumentID>[^,]*)',"
    "InsCode='(?P<InsCode>[^,]*)',"
    'BaseVol=(?P<BaseVol>[^,]*),'
    "EstimatedEPS='(?P<EstimatedEPS>[^,]*)',"
    'ZTitad=(?P<ZTitad>[^,]*),'
    "CIsin='(?P<CIsin>[^,]*)',"
    "LVal18AFC='(?P<LVal18AFC>[^,]*)',"
    "CSecVal='(?P<CSecVal>[^,]*)',"
    "PdrCotVal='(?P<PdrCotVal>[^,]*)',"
    "PClosing='(?P<PClosing>[^,]*)',"
    "PSGelStaMax='(?P<PSGelStaMax>[^,]*)',"
    "PSGelStaMin='(?P<PSGelStaMin>[^,]*)',"
    "Title='(?P<Title>[^,]*)',"
    "FaraDesc ='(?P<FaraDesc>[^,]*)',"
    "MinWeek='(?P<MinWeek>[^,]*)',"
    "MaxWeek='(?P<MaxWeek>[^,]*)',"
    "MinYear='(?P<MinYear>[^,]*)',"
    "MaxYear='(?P<MaxYear>[^,]*)',"
    "QTotTran5JAvg='(?P<QTotTran5JAvg>[^,]*)',"
    "SectorPE='(?P<SectorPE>[^,]*)',"
    "KAjCapValCpsIdx='(?P<KAjCapValCpsIdx>[^,]*)',"
    'PriceMin=(?P<PriceMin>[^,]*),'
    'PriceMax=(?P<PriceMax>[^,]*),'
    'PriceYesterday=(?P<PriceYesterday>[^;]*);'
    "ThemeCount='(?P<ThemeCount>[^;]*)';"
    "ContractSize='(?P<ContractSize>[^;]*)';"
    "NAV='(?P<NAV>[^;]*)';"
    "PSR='(?P<PSR>[^;]*)';"
).search
_TITLE_FULLMATCH = _rc(r"(.*?) \(.*?\) \- ([^']*)").fullmatch
_RELATED_COMPANIES = _rc(r'var RelatedCompanies=(\[.*\]);').search
_TRADE_HISTORY = _rc(r'var TradeHistory=(\[.*\]);').search
_STR_TO_NUM = _partial(_rc(rf"'{_F}'").sub, r'\1')


class _ClassProperty:
    def __init__(self, method):
        self.method = method

    def __get__(self, owner_self, owner_cls) -> _DataFrame:
        return self.method(owner_cls)

    def __set__(self, instance, value):
        pass


# note: this class is public through dataset module
class _LazyDS:
    l18s_to_l30_code: dict[str, tuple[str, str]]
    l30s_to_l18_code: dict[str, tuple[str, str]]
    path = Path(__file__).parent / 'dataset/dataset.csv'

    @_ClassProperty
    def df(cls) -> _DataFrame:
        df = _read_csv(
            cls.path,
            low_memory=False,
            lineterminator='\n',
            dtype='string',
            encoding='utf-8-sig',
            index_col='ins_code',
        )
        cls.df = df
        return df

    @classmethod
    def l30_code(cls, l18: str) -> tuple[str, str]:
        df = cls.df
        d = cls.l18s_to_l30_code = dict(
            zip(df['l18'], [*zip(df['l30'], df.index)])
        )
        g = cls.l30_code = d.get  # type: ignore
        return g(l18)  # type: ignore


class IntraDay(_TypedDict, total=False):
    general: dict
    thresholds: _DataFrame
    closings: _DataFrame
    candles: _DataFrame
    states: _DataFrame
    trades: _DataFrame
    holders: _DataFrame
    yesterday_holders: _DataFrame
    client_types: dict[str, int]
    best_limits: _DataFrame


class LiveData(_TypedDict, total=False):
    best_limits: _DataFrame
    market_state: MarketState
    nav: int
    nav_datatime: _jdatetime
    pc: int
    pf: int
    pl: int
    pmax: int
    pmin: int
    py: int
    status: str
    timestamp: _Ts
    tno: int
    tval: int
    tvol: int


class ETF(_TypedDict):
    insCode: str
    deven: int
    hEven: int
    pRedTran: float
    pSubTran: float
    iClose: int


class InstrumentState(_TypedDict):
    idn: int
    dEven: int
    hEven: int
    insCode: None
    lVal18AFC: None
    lVal30: None
    cEtaval: str
    realHeven: int
    underSupervision: int
    cEtavalTitle: str


class ClosingPriceInfo(_TypedDict):
    instrumentState: InstrumentState
    instrument: None
    lastHEven: int
    finalLastDate: int
    nvt: float
    mop: int
    thirtyDayClosingHistory: None
    priceChange: float
    priceMin: float
    priceMax: float
    priceYesterday: float
    priceFirst: float
    last: bool
    id: int
    insCode: str
    dEven: int
    hEven: int
    pClosing: float
    iClose: bool
    yClose: bool
    pDrCotVal: float
    zTotTran: float
    pRedTran: float
    qTotTran5J: float
    qTotCap: float


class ClientType(_TypedDict):
    buy_I_Volume: float
    buy_N_Volume: float
    buy_DDD_Volume: float
    buy_CountI: int
    buy_CountN: int
    buy_CountDDD: int
    sell_I_Volume: float
    sell_N_Volume: float
    sell_CountI: int
    sell_CountN: int


class Sector(_TypedDict):
    dEven: int
    cSecVal: str
    lSecVal: str


class SubSector(_TypedDict):
    dEven: int
    cSecVal: None
    cSoSecVal: int
    lSoSecVal: str


class Identity(_TypedDict):
    sector: Sector
    subSector: SubSector
    cValMne: str
    lVal18: str
    cSocCSAC: str
    lSoc30: str
    yMarNSC: str
    yVal: str
    insCode: str
    lVal30: str
    lVal18AFC: str
    flow: int
    cIsin: str
    zTitad: float
    baseVol: int
    instrumentID: str
    cgrValCot: str
    cComVal: str
    lastDate: int
    sourceID: int
    flowTitle: str
    cgrValCotTitle: str


class ShareHolder(_TypedDict):
    shareHolderID: int
    shareHolderName: str | None  # None: reserved code of ETFs
    cIsin: str
    dEven: int
    numberOfShares: float
    perOfShares: float
    change: int
    changeAmount: float
    shareHolderShareID: int


class Message(_TypedDict):
    tseMsgIdn: int
    dEven: int
    hEven: int
    tseTitle: str
    tseDesc: str
    flow: int


class Codal(_TypedDict):
    id: int
    symbol: str
    name: str
    title: str
    sentDateTime_Gregorian: str
    publishDateTime_Gregorian: str
    publishDateTime_DEven: int
    mainTableRowID: int
    hasHtmlReport: int
    hasExcelReport: int
    hasPDFReport: int
    hasXMLReport: int
    attachmentID: int
    contentType: int
    fileName: str
    fileExtension: str
    tracingNo: str


class ClientTypeOnDate(_TypedDict):
    recDate: int
    insCode: str
    buy_I_Volume: float
    buy_N_Volume: float
    buy_I_Value: float
    buy_N_Value: float
    buy_N_Count: int
    sell_I_Volume: float
    buy_I_Count: float
    sell_N_Volume: float
    sell_I_Value: float
    sell_N_Value: float
    sell_N_Count: int
    sell_I_Count: int


class Instrument:
    __slots__ = 'code', '_l18', '_l30', '_cisin', '_cs', '_isin'

    def __init__(
        self, code: str | int, l18: str | None = None, l30: str | None = None
    ):
        self.code = f'{code}'
        if l18 is not None:
            self._l18 = l18
        if l30 is not None:
            self._l30 = l30

    def __repr__(self):
        # not using self.l18 because it may require a web request
        try:
            return f'Instrument({self.code}, {self._l18!r})'
        except AttributeError:
            try:
                return f'Instrument({self.code}, l30={self._l30!r})'
            except AttributeError:
                return f'Instrument({self.code})'

    def __eq__(self, other):
        return self.code == other.code

    def __hash__(self):
        return int(self.code)

    @property
    async def l18(self) -> str:
        return await self._ds_or_info('l18')

    @property
    async def _arabic_l18(self) -> str:
        return (await self.l18).translate(_FARSI_NORM_REVERSED)

    @property
    async def l30(self) -> str:
        return await self._ds_or_info('l30')

    async def _ds_or_info(self, col: str) -> str:
        try:
            return getattr(self, f'_{col}')
        except AttributeError:
            pass
        try:
            v = _LazyDS.df.at[self.code, col]
        except KeyError:
            await self.info()
            return getattr(self, f'_{col}')
        else:
            setattr(self, f'_{col}', v)
            return v

    @property
    async def cisin(self) -> str:
        return await self._ds_or_info('cisin')

    @property
    async def isin(self) -> str:
        return await self._ds_or_info('isin')

    @property
    async def cs(self) -> str:
        try:
            return self._cs
        except AttributeError:
            await self.info()
        return self._cs

    @staticmethod
    async def from_l18(l18: str, /) -> 'Instrument':
        try:
            l30, ins_code = _LazyDS.l30_code(l18)
        except TypeError:  # cannot unpack non-iterable NoneType object
            return await Instrument.from_search(l18)
        return Instrument(ins_code, l18, l30)

    @staticmethod
    async def from_isin(isin: str, /) -> 'Instrument':
        """Create Instrument from ISIN.

        Raises KeyError if isin is not found in dataset.
        ISIN can be fetched using Instrument.info()['instrumentID'].
        """
        df = _LazyDS.df
        try:
            row = df[df['isin'] == isin].iloc[0]
        except KeyError as e:
            e.add_note('isin not found in dataset, try updating the dataset.')
            raise e
        inst = Instrument(row.name, row['l18'], row['l30'])  # type: ignore
        inst._isin = isin
        inst._cisin = row['cisin']
        return inst

    async def info(self) -> InstrumentInfo:
        j = await _api(f'Instrument/GetInstrumentInfo/{self.code}', fa=True)
        d = j['instrumentInfo']
        # cache for properties
        self._cs = d['sector']['cSecVal']
        self._cisin = d['cIsin']
        self._isin = d['instrumentID']
        self._l18 = d['lVal18AFC']
        self._l30 = d['lVal30']
        return j['instrumentInfo']

    async def trades(self) -> _DataFrame:
        j = await _api(f'Trade/GetTrade/{self.code}')
        df = _DataFrame(j['trade'], copy=False)
        return df

    async def codal(self, n=9) -> list[Codal]:
        j = await _api(
            f'Codal/GetPreparedDataByInsCode/{n}/{self.code}', fa=True
        )
        return j['preparedData']

    async def daily_closing_price(self, n=0) -> _DataFrame:
        """n is the number of days. Use 0 (default) to fetch all.

        result.index.is_monotonic_decreasing.
        """
        j = await _api(
            f'ClosingPrice/GetClosingPriceDailyList/{self.code}/{n}'
        )
        df = _DataFrame(j['closingPriceDaily'], copy=False)
        datetime = df['datetime'] = _to_datetime(
            df.pop('dEven').astype(str)
            + df.pop('hEven').astype(str).str.rjust(6, '0')
        )
        df['date'] = datetime.dt.normalize()
        df.set_index('date', inplace=True)
        return df

    async def closing_price_info(self) -> ClosingPriceInfo:
        j = await _api(
            f'ClosingPrice/GetClosingPriceInfo/{self.code}', fa=True
        )
        return j['closingPriceInfo']

    async def best_limits(self) -> _DataFrame:
        j = await _api(f'BestLimits/{self.code}')
        df = _DataFrame(j['bestLimits'], copy=False)
        return df

    async def client_type(self) -> ClientType:
        j = await _api(f'ClientType/GetClientType/{self.code}/1/0')
        return j['clientType']

    async def etf(self) -> ETF:
        """Return ETF data. (Includes redemption NAV and datetime of it).

        This method is only valid for ETFs.
        """
        j = await _api(f'Fund/GetETFByInsCode/{self.code}')
        return j['etf']

    async def related_companies(self, cs=None) -> dict[str, _DataFrame]:
        if cs is None:
            cs = await self.cs

        from tsetmc.general import related_companies

        return await related_companies(cs)

    async def page_data(
        self, general=True, trade_history=False, related_companies=False
    ) -> dict:
        """Return the static info found on instrument's page.

        :param general: parse general data including bvol, cisin, etc.
        :param trade_history: include trade_history in the result.
        :param related_companies: parse and include related_companies.

        For the meaning of yval column values see:
            https://members.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83

        For the meaning of other column names see:
            http://www.tsetmc.com/Site.aspx?ParTree=151713

        Alternative methods from the new API:
        - self.info
        - self.related_companies
        - self.daily_closing_price
        """
        _warn(
            '`Instrument.page_data` is deprecated; see its doc-string for alternatives',
            DeprecationWarning,
            stacklevel=2,
        )

        text = await _get_par_tree(f'151311&i={self.code}')
        if general:
            m: _Match[str] = _PAGE_VARS(text)  # type: ignore
            title_match: _Match[str] = _TITLE_FULLMATCH(m['Title'])  # type: ignore
            free_float = m['KAjCapValCpsIdx']
            eps = m['EstimatedEPS']
            sps = m['PSR']  # PSR = P/S = P / (sps: sales per share)
            sector_pe = m['SectorPE']
            l30 = self._l30 = title_match[1]
            l18 = self._l18 = m['LVal18AFC']
            cisin = self._cisin = m['CIsin']
            cs = self._cs = m['CSecVal']
            result = {
                'bvol': int(m['BaseVol']),
                'cisin': cisin,
                'cs': int(cs),
                'eps': int(eps) if eps else None,
                'sps': float(sps) if sps else None,
                'flow': int(m['Flow']),
                'free_float': int(free_float) if free_float else None,
                'group_code': m['CgrValCot'],
                'isin': m['InstrumentID'],
                'l18': l18,
                'l30': l30,
                'flow_name': title_match[2],
                'month_average_volume': int(m['QTotTran5JAvg']),
                'sector_name': m['LSecVal'],
                'sector_pe': float(sector_pe) if sector_pe else None,
                'tmax': float(m['PSGelStaMax']),
                'tmin': float(m['PSGelStaMin']),
                'week_max': float(m['MaxWeek']),
                'week_min': float(m['MinWeek']),
                'year_max': float(m['MaxYear']),
                'year_min': float(m['MinYear']),
                'z': int(m['ZTitad']),
            }
        else:
            result = {}
        if trade_history:
            m = _TRADE_HISTORY(text, pos=m.end())  # type: ignore
            th = _literal_eval(_STR_TO_NUM(m[1]))
            th = _DataFrame(
                th,
                copy=False,
                columns=(
                    'date',
                    'pc',
                    'py',
                    'pmin',
                    'pmax',
                    'tno',
                    'tvol',
                    'tval',
                ),
            )
            th['date'] = _to_datetime(th['date'], format='%Y%m%d')
            th.set_index('date', inplace=True)
            result['trade_history'] = th
        if related_companies:
            m = _RELATED_COMPANIES(text, m.end())  # type: ignore
            result['related_companies'] = [
                Instrument(code, l18, l30)
                for (code, l18, l30) in _literal_eval(_STR_TO_NUM(m[1]))
            ]
        return result

    async def live_data(
        self, general=True, best_limits=False, market_state=False
    ) -> LiveData:
        """Return live data price/order data using instinfodata.aspx module.

        :keyword best_limits: parse best_limits and include related values.
        :keyword market_state: parse values related to market state.

        Alternatives from the new API:
        - `Instrument.closing_price_info`
        - `Instrument.best_limits`
        - `Instrument.etf`
        - `general.market_overview`
        """
        _warn(
            '`Instrument.live_data` is deprecated; see its doc-string for alternatives',
            DeprecationWarning,
            stacklevel=2,
        )
        # apparently, http://www.tsetmc.com/tsev2/data/instinfodata.aspx?i=...
        # and http://www.tsetmc.com/tsev2/data/instinfofast.aspx?i=...
        # return the same response.
        text = await _get_data(
            'instinfodata.aspx'
            # &e=1 parameter is required to get cancel NAV for ETFs but it
            # seems to be ignored if no NAV is defined for the instrument.
            # &c= is normally set to CSecVal, but does not seem to be required.
            # CSecVal is industry group code:
            # http://redirectcdn.tsetmc.com/Site.aspx?ParTree=111411111B&LnkIdn=107
            f'?i={self.code}&c=&e=1',
            fa=True,
        )
        # the _s are unknown
        try:
            (
                price_info,
                index_info,
                orders_info,
                _,
                _,
                _,
                group_info,
                _,
                _,
            ) = text.split(';')
        except ValueError:
            _warning(text)  # usually means the service is unavailable
            raise
        if general:
            result = _parse_price_info(price_info)
        else:
            result = {}
        if market_state:
            if index_info:  # sometimes index_info is an empty string
                result['market_state'] = _parse_market_state(index_info)
        if best_limits:
            result['best_limits'] = _csv2df(
                _StringIO(orders_info),
                sep='@',
                names=('zd', 'qd', 'pd', 'po', 'qo', 'zo'),
                lineterminator=',',
            )
        return result  # type: ignore

    async def trade_history(self, top: int, all_=False) -> _DataFrame:
        """Get history of pmax, pmin, pc, pl, pf, py, tval, tvol, and tno.

        :param top: number of top rows (days) to return
        :param all_: include dates with no trade

        Use :meth:`Instrument.on_date(<date>).trades` for intraday trades.
        """
        _warn(
            '`Instrument.trade_history` is deprecated; use `Instrument.daily_closing_price` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        content = await _get_data(
            f'InstTradeHistory.aspx?i={self.code}&Top={top}&A={all_:d}'
        )
        df = _csv2df(
            _BytesIO(content),
            sep='@',
            names=(
                'date',
                'pmax',
                'pmin',
                'pc',
                'pl',
                'pf',
                'py',
                'tval',
                'tvol',
                'tno',
            ),
            index_col='date',
            parse_dates=True,
        )
        return df

    async def price_history(self, adjusted: bool = True) -> _DataFrame:
        # As far as I can thll the new tsetmc site does not have any
        # API for adjusted price history, but see self.price_adjustments.
        content = await _get(
            f'https://members.tsetmc.com/tsev2/chart/data/Financial.aspx?i={self.code}&t=ph&a={adjusted:d}'
        )
        df = _csv2df(
            _BytesIO(content),
            names=('date', 'pmax', 'pmin', 'pf', 'pl', 'tvol', 'pc'),
            index_col='date',
            parse_dates=True,
        )
        return df

    async def client_type_history_old(self) -> _DataFrame:
        """Get daily natural/legal history.

        In column names `n_` prefix stands for natural and `l_` for legal.

        This method returns the information available at the "حقیقی-حقوقی" tab
        of the instrument. It uses the `clienttype.aspx` module of
        old.tsetmc.com.

        This method may stop working in the future if the old tsetmc site
        shuts down.

        See also:
            :meth:`Instrument.client_type_history`
        """
        _warn(
            '`Instrument.client_type_history_old` is deprecated; use `Instrument.client_type_history` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return _csv2df(
            _BytesIO(await _get_data(f'clienttype.aspx?i={self.code}')),
            names=(
                'date',
                'n_buy_count',
                'l_buy_count',
                'n_sell_count',
                'l_sell_count',
                'n_buy_volume',
                'l_buy_volume',
                'n_sell_volume',
                'l_sell_volume',
                'n_buy_value',
                'l_buy_value',
                'n_sell_value',
                'l_sell_value',
            ),
            index_col='date',
            parse_dates=True,
            dtype='int64',
        )

    @_overload
    async def client_type_history(self, date: None = None) -> _DataFrame: ...
    @_overload
    async def client_type_history(
        self, date: int | str
    ) -> ClientTypeOnDate: ...
    async def client_type_history(
        self, date: int | str | None = None
    ) -> _DataFrame | ClientTypeOnDate:
        """Return natural/legal client type history.

        :param date: Gregorian date in YYYYMMDD format. If None, return the
            full history as a DataFrame. Otherwise, return the data for that
            specific date as a dict.

        Legal persoans are indicated with `N`, natural persons with `I`.

        See also:
            :meth:`Instrument.client_type`
        """
        if date is None:
            # the following api call returns a result that is equivalent to
            # j = (await _api(f'ClientType/GetClientType/{self.code}/1/0'))[
            #     'clientType']
            j = await _api(f'ClientType/GetClientTypeHistory/{self.code}')
            return _DataFrame(j['clientType'], copy=False)

        j = await _api(f'ClientType/GetClientTypeHistory/{self.code}/{date}')
        return j['clientType']

    async def identification(self) -> dict:
        """Return the information available in the identification (شناسه) tab.

        Related API descriptions:
            https://cdn.tsetmc.com/Site.aspx?ParTree=1114111118&LnkIdn=83
            http://en.tsetmc.com/Site.aspx?ParTree=111411111Z

        This method uses old.tsetmc.com.
        For the new API use `Instrument.identity`.
        """
        _warn(
            '`Instrument.identification` is deprecated; use `Instrument.identity` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        text = await _get_par_tree(f'15131M&i={self.code}')
        df = _html_to_df(text)
        return dict(zip(df[0], df[1]))

    async def identity(self) -> Identity:
        j = await _api(
            f'Instrument/GetInstrumentIdentity/{self.code}', fa=True
        )
        return j['instrumentIdentity']

    async def introduction(self) -> dict[str, str]:
        """Return the information available in introduction (معرفی) tab."""
        _warn(
            '`Instrument.introduction` is deprecated; use `Instrument.publisher` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        text = await _get_par_tree(f'15131V&s={await self._arabic_l18}')
        df = _html_to_df(text)
        return dict(zip(df[0].str.removesuffix(' :'), df[1]))

    async def publisher(self) -> dict:
        j = await _api(
            f'Codal/GetCodalPublisherBySymbol/{await self._arabic_l18}',
            fa=True,
        )
        return j['codalPublisher']

    @staticmethod
    async def from_search(s: str) -> 'Instrument':
        """`Search for `s` and return the first result as an Instrument."""
        d = (await search(s))[0]
        return Instrument(d['insCode'], d['lVal18AFC'], d['lVal30'])

    async def share_holders(self) -> list[ShareHolder]:
        """Return a list of the current major holders of this instrument."""
        j = await _api(
            f'Shareholder/GetInstrumentShareHolderLast/{self.code}', fa=True
        )
        return j['shareHolder']

    async def share_holder_history(
        self,
        share_holder_id: int,
        days: int = 90,
    ) -> _DataFrame:
        """Return history of share changes of a particular holder.

        Obtain `share_holder_id` from `self.share_holders`.
        To get other companies of the same shareholder use the
        `share_holder_companies` function.
        """
        j = await _api(
            f'Shareholder/GetShareHolderHistory/{self.code}/{share_holder_id}/{days}',
            fa=True,
        )
        df = _DataFrame(j['shareHolder'], copy=False)
        df['dEven'] = _to_datetime(df['dEven'], format='%Y%m%d')
        df.set_index('dEven', inplace=True)
        return df

    async def holders(self, cisin=None) -> _DataFrame:
        """Get list of current major unit/shareholders.

        If `cisin` is not provided, it will be fetched using a web request.

        See also:
            :meth:`Instrument.on_date(<date>).holders`
        """
        _warn(
            '`Instrument.holders` is deprecated; use `Instrument.share_holders` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        if cisin is None:
            cisin = await self.cisin
        text = await _get_par_tree(f'15131T&c={cisin}')
        df = _html_to_df(text)
        df.drop(columns='Unnamed: 4', inplace=True)
        # todo: use separate columns
        df.columns = ['holder', 'shares/units', '%', 'change']
        df['id_cisin'] = _findall(r"ShowShareHolder\('([^']*)'\)", text)
        if df['change'].dtype == 'string':
            _numerize(df, ('change',), 'Int64')
        _numerize(df, ('shares/units',), 'int64')
        return df

    @_overload
    @staticmethod
    async def holder(
        id_cisin, history: _Literal[True], other_holdings: _Literal[True]
    ) -> tuple[_DataFrame, _DataFrame]: ...

    @_overload
    @staticmethod
    async def holder(
        id_cisin, history: bool = True, other_holdings: bool = False
    ) -> _DataFrame: ...

    @staticmethod
    async def holder(
        id_cisin, history=True, other_holdings=False
    ) -> _DataFrame | tuple[_DataFrame, _DataFrame]:
        """Return history/other holdings for the given holder id_cisin.

        `id_cisin` is usually obtained using `self.holders`.

        If both `history` and `other_holdings` are True, then a tuple of
        DataFrames will be returned.
        """
        _warn(
            '`Instrument.holder` is deprecated; use `Instrument.share_holder_history` or `instruments.share_holder_companies` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        text = await _get_data(f'ShareHolder.aspx?i={id_cisin}', fa=True)
        hist, _, oth = text.partition('#')

        def history_df() -> _DataFrame:
            return _csv2df(
                _StringIO(hist),
                names=('date', 'shares'),
                dtype='int64',
                index_col='date',
                parse_dates=True,
            )

        def other_holdings_df() -> _DataFrame:
            return _csv2df(
                _StringIO(oth),
                names=('ins_code', 'name', 'shares', 'percent'),
                index_col='ins_code',
            )

        if history and other_holdings:
            return history_df(), other_holdings_df()
        elif history:
            return history_df()
        else:
            return other_holdings_df()

    async def adjustments(self) -> _DataFrame:
        _warn(
            '`Instrument.adjustments` is deprecated; use `Instrument.price_adjustments` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        content = await _get_par_tree(f'15131G&i={self.code}', fa=False)
        df = _html_to_df(content.decode())
        df.columns = ('date', 'adj_pc', 'pc')
        df['date'] = [_j_ymd_parse(d) for d in df['date']]
        return df

    async def price_adjustments(self) -> _DataFrame:
        j = await _api(f'ClosingPrice/GetPriceAdjustList/{self.code}')
        df = _DataFrame(j['priceAdjust'], copy=False)
        df['dEven'] = _to_datetime(df['dEven'], format='%Y%m%d')
        df.set_index('dEven', inplace=True)
        return df

    async def messages(self) -> list[Message]:
        """See also: ``general.messages`` and ``general.search_messages``."""
        j = await _api(f'Msg/GetMsgByInsCode/{self.code}', fa=True)
        return j['msg']

    async def ombud_messages(self) -> _DataFrame:
        _warn(
            '`Instrument.ombud_messages` is deprecated; use `Instrument.messages` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return _parse_ombud_messages(
            await _get_par_tree(f'15131W&i={self.code}')
        )

    async def dps_history(self) -> _DataFrame:
        """Get DPS history.

        :raises pandas.errors.EmptyDataError: No columns to parse from file
        """
        # Note: Currently does not have an _api equivalent. The DPS tab is
        # non-functional in the new tsetmc website.
        content = await _get_data(f'DPSData.aspx?s={await self._arabic_l18}')
        df = _csv2df(
            _BytesIO(content),
            header=None,
            sep='@',
        )
        cols = [
            'publish_date',
            'meeting_date',
            'fiscal_year',
            'profit_or_loss_after_tax',
            'distributable_profit',
            'accumulated_profit_at_the_end_of_the_period',
            'cash_earnings_per_share',
        ]
        df.columns = cols
        for col in cols[:3]:
            df[col] = [_jstrptime(d, '%Y/%m/%d') for d in df[col]]
        return df

    def on_date(self, date: int | str) -> 'InstrumentOnDate':
        """Return an object resembling Instrument on a specific date.

        :param date: Gregorian date in YYYYMMDD format.

        :note: methods of InstrumentOnDate are rate-limited.
            Avoid making simultaneous calls.
        """
        return InstrumentOnDate(_inst=self, _date=date)


class ClosingPrice(_TypedDict):
    priceChange: float
    priceMin: float
    priceMax: float
    priceYesterday: float
    priceFirst: float
    last: bool
    id: int
    insCode: str
    dEven: int
    hEven: int
    pClosing: float
    iClose: bool
    yClose: bool
    pDrCotVal: float
    zTotTran: float
    qTotTran5J: float
    qTotCap: float


class InstrumentOnDate:
    __slots__ = 'date', 'code', 'inst'

    def __init__(self, /, *, _inst: Instrument, _date: int | str):
        """Return an object resembling Instrument on a specific date.

        :param _inst: Instrument
        :param _date: Gregorian date in YYYYMMDD format.

        Users should not instantiate this class directly, but use
        :meth:`Instrument.on_date` instead.

        :note: methods of InstrumentOnDate are rate-limited.
            Avoid making simultaneous calls.
        """
        self.date = _date
        self.inst = _inst
        self.code = _inst.code

    async def closing_price(self) -> ClosingPrice:
        """Return general closing price info.

        Result dict has the following keys: {
        'priceChange', 'priceMin', 'priceMax', 'priceYesterday', 'priceFirst',
        'last', 'id', 'insCode', 'dEven', 'hEven', 'pClosing', 'iClose',
        'yClose', 'pDrCotVal', 'zTotTran', 'qTotTran5J', 'qTotCap'}
        """
        j = await _api(
            f'ClosingPrice/GetClosingPriceDaily/{self.code}/{self.date}'
        )
        return j['closingPriceDaily']

    async def closing_price_history(self) -> _DataFrame:
        """Get intraday closing price history."""
        j = await _api(
            f'ClosingPrice/GetClosingPriceHistory/{self.code}/{self.date}'
        )
        return _DataFrame(j['closingPriceHistory'], copy=False)

    async def states(self) -> _DataFrame:
        """Get intraday instrument states.

        http://www.tsetmc.com/Site.aspx?ParTree=111411111Y&LnkIdn=833
        """
        j = await _api(
            f'MarketData/GetInstrumentState/{self.code}/{self.date}'
        )
        return _DataFrame(j['instrumentState'], copy=False)

    async def client_type(self) -> ClientTypeOnDate:
        return await self.inst.client_type_history(self.date)

    async def client_types(self) -> ClientTypeOnDate:
        _warn(
            '`InstrumentOnDate.client_types` is deprecated; use `InstrumentOnDate.client_type` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return await self.client_type()

    async def holders(self) -> _DataFrame:
        """Return share/unit holders for a specific date and a day before that.

        See also:
            :meth:`Instrument.holders` which returns the list of current
            holders.
        """
        j = await _api(f'Shareholder/{self.code}/{self.date}')
        return _DataFrame(j['shareShareholder'], copy=False)

    async def best_limits(self) -> _DataFrame:
        """Get intraday best limits history."""
        j = await _api(f'BestLimits/{self.code}/{self.date}')
        return _DataFrame(j['bestLimitsHistory'], copy=False)

    async def trades(self) -> _DataFrame:
        """Get intraday trades.

        See also:
           :meth:`Instrument.trade_history`
        """
        # todo: true vs false
        j = await _api(f'Trade/GetTradeHistory/{self.code}/{self.date}/true')
        return _DataFrame(j['tradeHistory'], copy=False)

    async def static_thresholds(self) -> _DataFrame:
        """Get intraday static thresholds."""
        j = await _api(
            f'MarketData/GetStaticThreshold/{self.code}/{self.date}'
        )
        return _DataFrame(j['staticThreshold'], copy=False)

    async def data(self) -> dict:
        """Get general info about the instrument on the specific date.

        The returned dict contains the following keys: {
        'insCode', 'lVal30', 'lVal18AFC', 'flow', 'cIsin', 'zTitad', 'baseVol',
        'instrumentID', 'cgrValCot', 'cComVal', 'lastDate', 'sourceID',
        'flowTitle', 'cgrValCotTitle'}
        """
        j = await _api(
            f'Instrument/GetInstrumentHistory/{self.code}/{self.date}'
        )
        return j['instrumentHistory']


async def price_adjustments(flow: _FlowType) -> _DataFrame:
    """Get price adjustments for a particular flow.

    Related APIs:
        http://cdn.tsetmc.com/Site.aspx?ParTree=1114111124&LnkIdn=843
    """
    text = await _get_par_tree(f'151319&Flow={flow}')
    df = _html_to_df(text)
    df.columns = ('l18', 'l30', 'date', 'adj_pc', 'pc')
    df['date'] = [_j_ymd_parse(d) for d in df['date']]
    return df


async def old_search(skey: str, /) -> _DataFrame:
    """`skey` (search key) is usually part of the l18 or l30."""
    return _csv2df(
        _StringIO(await _get_data('search.aspx?skey=' + skey, fa=True)),
        header=None,
        # Another terminology would be: {
        #   ins_code: round_lot,
        #   retail: odd_lot,
        #   compensation: buyback,
        #   wholesale: block,
        # } see: https://www.sena.ir/news/77488/
        names=(
            'l18',
            'l30',
            'ins_code',
            'retail',
            'compensation',
            'wholesale',
            '_unknown1',
            '_unknown2',
            '_unknown3',
            '_unknown4',
            '_unknown5',
        ),
    )


class Search(_TypedDict):
    insCode2: str
    insCode3: str
    insCode4: str
    insCode: str
    lVal30: str
    lVal18AFC: str
    flow: int
    cIsin: None
    zTitad: float
    baseVol: int
    instrumentID: None
    cgrValCot: str
    cComVal: None
    lastDate: int
    sourceID: int
    flowTitle: str
    cgrValCotTitle: str


async def search(s: str, /) -> list[Search]:
    r = await _api(f'Instrument/GetInstrumentSearch/{s}', fa=True)
    return r['instrumentSearch']


class InstrumentType(_TypedDict, total=False):
    cValMne: None
    lVal18: None
    cSocCSAC: None
    lSoc30: None
    yMarNSC: None
    yVal: None
    insCode: str
    lVal30: str
    lVal18AFC: str
    flow: int
    cIsin: None
    zTitad: float
    baseVol: int
    instrumentID: None
    cgrValCot: None
    cComVal: None
    lastDate: int
    sourceID: int
    flowTitle: None
    cgrValCotTitle: None


class ShareHolderCompany(_TypedDict):
    instrument: InstrumentType
    numberOfShares: float
    perOfShares: float


async def share_holder_companies(
    share_holder_share_id: int | str, /
) -> list[ShareHolderCompany]:
    r = await _api(
        f'Shareholder/GetShareHolderCompanyList/{share_holder_share_id}',
        fa=True,
    )
    return r['shareHolderShare']


def _parse_price_info(price_info):
    price_parts = price_info.split(',')

    if len(price_parts) == 17:
        # this is unexpected according to js source
        assert price_parts[14] == price_parts[10]
        del price_parts[14]

    (  # see dev/tsetmc_source_files/Instinfo.js
        time,  # 0
        status,  # 1
        pl,  # 2
        pc,  # 3
        pf,  # 4
        py,  # 5
        pmin,  # 6
        pmax,  # 7
        tno,  # 8
        tvol,  # 9
        tval,  # 10
        _,  # 11 is used to add css styles for pc/pl, not important.
        info_datetime_date,  # 12
        last_info_time,  # 13
        nav_datetime,  # 14
        nav,  # 15
    ) = price_parts

    result = {
        'time': time,
        'status': status,
        'timestamp': _Ts(f'{info_datetime_date}{last_info_time:>06}'),
        'pl': int(pl),
        'pc': int(pc),
        'pf': int(pf),
        'py': int(py),
        'pmin': int(pmin),
        'pmax': int(pmax),
        'tno': int(tno),
        'tvol': int(tvol),
        'tval': int(tval),
    }

    if nav:
        result['nav'] = int(nav)
        try:
            _jdatetime(1400, 12, 1, 16, 30, 0)
            _fullmatch(r'(\d+)/(\d+)/(\d+) (\d+):(\d+):(\d+)', nav_datetime)
            result['nav_datetime'] = _jstrptime(
                nav_datetime, '%Y/%m/%d %H:%M:%S'
            )
        except ValueError:  # could be invalid day of the month
            result['nav_datetime'] = nav_datetime

    return result


def _parse_ombud_messages(text) -> _DataFrame:
    headers = _findall(r'<th>(.+?)</th>', text)
    dates = _findall(r"<th class='ltr'>(.+?)</th>", text)
    descriptions = _findall(r'<td colspan="2">(.+?)<hr />\s*</td>', text)
    df = _DataFrame(
        {'header': headers, 'date': dates, 'description': descriptions},
        dtype='string',
        copy=False,
    )
    if dates:  # pandas cannot do ('14' + df['date']) on empty dates
        df['date'] = [
            _jstrptime(d, format='%Y/%m/%d %H:%M') for d in ('14' + df['date'])
        ]
    else:
        df['date'] = df['date'].astype(object)
    return df



from pandas import (
    json_normalize as _json_normalize,
    to_datetime as _to_datetime,
)

from tsetmc import _api, _DataFrame
from tsetmc.instruments import InstrumentInfo


async def last_state(i=1, /) -> _DataFrame:
    j = await _api(f'Index/GetIndexB1LastAll/All/{i}', fa=True)
    return _DataFrame(j['indexB1'], copy=False)


class Index:
    __slots__ = 'code'

    def __init__(self, code: str | int):
        self.code = code

    async def info(self) -> InstrumentInfo:
        j = await _api(f'Instrument/GetInstrumentInfo/{self.code}', fa=True)
        return j['instrumentInfo']

    async def last_day_history(self) -> _DataFrame:
        j = await _api(f'Index/GetIndexB1LastDay/{self.code}', fa=True)
        df = _DataFrame(j['indexB1'])
        dt = _to_datetime(
            df.pop('dEven').astype(str) + df.pop('hEven').astype(str),
            format='%Y%m%d%H%M%S',
        )
        df.set_index(dt, inplace=True)
        return df

    async def history(self) -> _DataFrame:
        j = await _api(f'Index/GetIndexB2History/{self.code}', fa=True)
        df = _DataFrame(j['indexB2'])
        date = _to_datetime(df.pop('dEven'), format='%Y%m%d')
        df.set_index(date, inplace=True)
        return df

    async def companies(self) -> dict[str, _DataFrame]:
        j = await _api(f'ClosingPrice/GetIndexCompany/{self.code}', fa=True)
        j['indexCompany'] = _json_normalize(j['indexCompany'])
        j['relatedCompanyThirtyDayHistory'] = _DataFrame(
            j['relatedCompanyThirtyDayHistory']
        )
        return j



__version__ = '0.68.2'
from enum import StrEnum as _StrEnum
from functools import partial as _partial
from json import JSONDecodeError, loads
from logging import getLogger as _getLogger
from re import Match as _Match, compile as _rc
from typing import (
    Literal as _Literal,
    TypedDict as _TypedDict,
    overload as _overload,
)

from aiohutils.session import SessionManager
from jdatetime import datetime as _jdatetime
from pandas import (
    NA as _NA,
    DataFrame as _DataFrame,
    options as _o,
    read_csv as _read_csv,
)

_logger = _getLogger(__name__)
_o.mode.copy_on_write = True
_o.future.infer_string = True  # type: ignore
_o.future.no_silent_downcasting = True  # type: ignore
_csv2df = _partial(_read_csv, low_memory=False, engine='c', lineterminator=';')
_F = r'(-?\d+(?:\.\d+)?)'  # float pattern


_INDEX_CHANGE_MATCH = _rc(rf'<div[^>]*>(\()?{_F}\)?</div>(?: {_F}%)?').match
_INDEX_TIMESTAMP_MATCH = _rc(r'(\d\d)/(\d+)/(\d+) (\d\d):(\d\d):(\d\d)').match


_jstrptime = _jdatetime.strptime


class MarketState(_TypedDict, total=False):
    datetime: _jdatetime
    tse_status: str
    tse_index: float
    tse_index_change: float | None
    tse_index_change_percent: float | None
    tse_tvol: float
    tse_tval: float
    tse_tno: float
    fb_status: str
    fb_tvol: float
    fb_tval: float
    fb_tno: int
    tse_value: float | None
    derivatives_status: str
    derivatives_tval: float
    derivatives_tvol: float
    derivatives_tno: int


class Eps(_TypedDict):
    epsValue: None
    estimatedEPS: str | None
    sectorPE: float
    psr: float


class Sector(_TypedDict):
    dEven: int
    cSecVal: str
    lSecVal: str


class StaticThreshold(_TypedDict):
    insCode: None
    dEven: int
    hEven: int
    psGelStaMax: float
    psGelStaMin: float


class InstrumentInfo(_TypedDict):
    eps: Eps
    sector: Sector
    staticThreshold: StaticThreshold
    minWeek: float
    maxWeek: float
    minYear: float
    maxYear: float
    qTotTran5JAvg: float
    kAjCapValCpsIdx: str
    dEven: int
    topInst: int
    faraDesc: str
    contractSize: int
    nav: float
    underSupervision: int
    cValMne: None
    lVal18: str
    cSocCSAC: None
    lSoc30: None
    yMarNSC: None
    yVal: str
    insCode: str
    lVal30: str
    lVal18AFC: str
    flow: int
    cIsin: str
    zTitad: float
    baseVol: int
    instrumentID: str
    cgrValCot: str
    cComVal: str
    lastDate: int
    sourceID: int
    flowTitle: str
    cgrValCotTitle: str


class Flow(_StrEnum):
    # Most of these values were obtained by checking different tabs on the
    # main page of tsetmc.com
    GENERAL = '0'  # Contains both BOURSE and OTC
    BOURSE = '1'
    OTC = '2'  # Over-the-counter, فرابورس
    FUTURES = '3'
    DERIVATIVES = '3'
    DEBT = '3'
    UTP = '4'  # Unlisted Trading Privileges, بازار پایه
    ENERGY = '6'
    MERCANTILE = '7'  # commodities
    ETFS = '18'
    PROFESSIONAL = '19'  # professional market


FlowType = int | str | Flow


def _parse_market_state(s: str) -> MarketState:
    splits = s.split(',')
    # Sometimes splits start with two timestamps. it is not clear what it is, the
    # UpdateFastView function in market_watch.js is not designed to handle this
    # situation. It could be that the first timestamp is last transaction time
    # and the second one is server time.
    if len(splits) == 18:
        splits.pop(0)
    try:
        (
            datetime,
            tse_status,
            tse_index,
            tse_index_change,
            tse_value,
            tse_tvol,
            tse_tval,
            tse_tno,
            fb_status,
            fb_tvol,
            fb_tval,
            fb_tno,
            derivatives_status,
            derivatives_tvol,
            derivatives_tval,
            derivatives_tno,
            _,
        ) = splits
    except ValueError as e:
        e.add_note(f'{s=}')
        raise
    if tse_index_change:  # can be '' before market start
        index_change_match: _Match = _INDEX_CHANGE_MATCH(tse_index_change)  # type: ignore
        tse_index_change = float(index_change_match[2])
        if index_change_match[1] is not None:  # negative value
            tse_index_change *= -1
        if (m3 := index_change_match[3]) is not None:
            tse_index_change_percent = float(m3)
        else:
            tse_index_change_percent = None
    else:
        tse_index_change_percent = None
        tse_index_change = None
    timestamp_match: _Match = _INDEX_TIMESTAMP_MATCH(datetime)  # type: ignore
    return {
        'datetime': _jdatetime(
            1400 + int(timestamp_match[1]),
            int(timestamp_match[2]),
            int(timestamp_match[3]),
            int(timestamp_match[4]),
            int(timestamp_match[5]),
            int(timestamp_match[6]),
        ),
        'tse_status': tse_status,
        'tse_index': float(tse_index),
        'tse_index_change': tse_index_change,
        'tse_tvol': float(tse_tvol),
        'tse_tval': float(tse_tval),
        'tse_tno': float(tse_tno),
        'fb_status': fb_status,
        'fb_tvol': float(fb_tvol),
        'fb_tval': float(fb_tval),
        'fb_tno': int(fb_tno),
        'derivatives_status': derivatives_status,
        'derivatives_tvol': float(derivatives_tvol),
        'derivatives_tval': float(derivatives_tval),
        'derivatives_tno': int(derivatives_tno),
        'tse_index_change_percent': tse_index_change_percent,
        'tse_value': float(tse_value) if tse_value else None,
    }


session_manager = SessionManager()


_FARSI_NORM = ''.maketrans('يك', 'یک')
_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:100.0) Gecko/20100101 Firefox/100.0'
}


# this function should only be called from _get below
async def _session_get(url: str) -> bytes:
    r = await session_manager.get(url, headers=_HEADERS)
    return await r.read()


_LAST_CONTENT: bytes


@_overload
async def _get(url: str, *, fa: _Literal[True]) -> str: ...


@_overload
async def _get(url: str, *, fa=False) -> bytes: ...


async def _get(url: str, *, fa: _Literal[True, False] = False) -> str | bytes:
    global _LAST_CONTENT
    _LAST_CONTENT = content = await _session_get(url)
    if fa is True:
        content = content.decode().translate(_FARSI_NORM)
    return content


_DOMAIN = 'http://old.tsetmc.com/'
_MEMBERS = 'http://members.tsetmc.com/'
# API does not work on www domain
_API = 'https://cdn.tsetmc.com/api/'


@_overload
async def _get_data(path: str, *, fa: _Literal[True]) -> str: ...
@_overload
async def _get_data(path: str, *, fa=False) -> bytes: ...
async def _get_data(path: str, *, fa=False) -> str | bytes:
    return await _get(f'{_DOMAIN}tsev2/data/' + path, fa=fa)


@_overload
async def _get_par_tree(path: str, *, fa: _Literal[False]) -> bytes: ...
@_overload
async def _get_par_tree(path: str, *, fa=True) -> str: ...
async def _get_par_tree(path: str, *, fa=True) -> str | bytes:
    return await _get(f'{_DOMAIN}Loader.aspx?ParTree={path}', fa=fa)


@_overload
async def _mem_par_tree(path: str, *, fa: _Literal[False]) -> bytes: ...
@_overload
async def _mem_par_tree(path: str, *, fa=True) -> str: ...
async def _mem_par_tree(path: str, *, fa=True) -> str | bytes:
    return await _get(f'{_MEMBERS}Loader.aspx?ParTree={path}', fa=fa)


async def _api(path: str, *, fa=False):
    content = await _get(f'{_API}{path}', fa=fa)
    try:
        return loads(content)
    except JSONDecodeError:
        _logger.error(f'url={_API}{path}\n{content=}')
        raise


def _numerize(
    df: _DataFrame,
    cols: tuple[str, ...],
    astype: _Literal['float64', 'int64', 'Int64'] = 'float64',
    comma=False,
):
    for col in cols:
        c = df[col]
        if comma is True:
            c = c.str.replace(',', '')
        # https://stackoverflow.com/a/39684629/2705757
        df[col] = (
            c.replace(r' [KMB]$', '', regex=True).astype(astype)
        ) * c.str.extract(r'[\d\. ]+([KMBT]+)', expand=False).map(
            {
                _NA: 1,
                'K': 10**3,
                'M': 10**6,
                'B': 10**9,
                'T': 10**12,
            }
        ).astype(astype)


def _save_last_content(msg: str, /):
    if (
        'دسترسی شما بدلیل مسائل امنیتی مسدود شده است.'.encode()
        in _LAST_CONTENT
    ):
        _logger.error('Your access has been blocked due to security issues.')
        return
    if b'General Error Detected' in _LAST_CONTENT:
        _logger.error('General Error Detected.')
        return

    file = __file__.removesuffix('__init__.py') + '~last_response.html'
    with open(file, 'wb') as f:
        f.write(_LAST_CONTENT)
    _logger.error(f'{msg}; _LAST_CONTENT saved in {file}', stacklevel=2)



from enum import StrEnum as _StrEnum

from pandas import (
    json_normalize as _json_normalize,
)

from tsetmc import (
    Flow as _Flow,
    FlowType as _FlowType,
    _api,
    _DataFrame,
)


class FundType(_StrEnum):
    STOCK = '6'
    MIXED = '7'
    FIXED = '4'
    MARKET_MAKING = '11'
    VC = '12'
    REIT = '14'
    PROJECT = '13'
    COMMODITY = '5'
    PRIVATE = '16'
    FUND = '17'


async def funds(type_: FundType | int | str, /) -> _DataFrame:
    """tsetmc.com > صندوق های سرمایه گذاری"""
    j = await _api(f'Fund/GetFunds/{type_}')
    return _DataFrame(j['funds'])


async def commodity_etfs(
    *, flow: _FlowType = _Flow.MERCANTILE, top: int | str = '9999'
) -> _DataFrame:
    """tsetmc.com > بورس کالا > صندوق های قابل معامله"""
    j = await _api(
        f'ClosingPrice/GetTradeTop/CommodityFund/{flow}/{top}', fa=True
    )
    return _json_normalize(j['tradeTop'])


async def etfs_with_most_price_decrease(
    *, flow: _FlowType = _Flow.BOURSE, top: int | str = '9999'
) -> _DataFrame:
    """tsetmc.com > بورس اوراق بهادار تهران > صندوق های قابل معامله > بیشترین کاهش قیمت"""
    j = await _api(
        f'ClosingPrice/GetTradeTop/PClosingBtmETF/{flow}/{top}', fa=True
    )
    return _json_normalize(j['tradeTop'])


async def etfs_with_most_price_increase(
    *, flow: _FlowType = _Flow.BOURSE, top: int | str = '9999'
) -> _DataFrame:
    """tsetmc.com > بورس اوراق بهادار تهران > صندوق های قابل معامله > بیشترین افزایش قیمت"""
    j = await _api(
        f'ClosingPrice/GetTradeTop/PClosingTopETF/{flow}/{top}', fa=True
    )
    return _json_normalize(j['tradeTop'])


async def most_traded_etfs(
    *, flow: _FlowType = _Flow.BOURSE, top: int | str = '9999'
) -> _DataFrame:
    """tsetmc.com > بورس اوراق بهادار تهران > صندوق های قابل معامله > بیشترین حجم معامله"""
    j = await _api(
        f'ClosingPrice/GetTradeTop/MostTradedETF/{flow}/{top}', fa=True
    )
    return _json_normalize(j['tradeTop'])


async def etfs(
    *, flow: _FlowType = _Flow.BOURSE, top: int | str = '9999'
) -> _DataFrame:
    """tsetmc.com > بورس اوراق بهادار تهران > صندوق های قابل معامله > معاملات صندوق های قابل معامله"""
    j = await _api(f'ClosingPrice/GetTradeTop/ETF/{flow}/{top}', fa=True)
    return _json_normalize(j['tradeTop'])



"""
https://tsetmc.com/StaticContent/WebServiceHelp
"""

from aiohutils.pd import html_to_df as _html_to_df
from lxml.html import fromstring as _html

from tsetmc import _api


async def _static_content(key: str):
    j = await _api(f'StaticData/GetStaticContent/WS-{key}')
    return j['staticContent'][0]['content']


async def client_type() -> dict:
    """https://tsetmc.com/StaticContent/WS-ClientType"""
    text = await _static_content('ClientType')
    out = _html_to_df(text, 2)
    return {
        'output': out,
    }


async def instrument_filter_by_date() -> dict:
    """https://tsetmc.com/StaticContent/WS-InstrumentFilterByDate"""
    text = await _static_content('InstrumentFilterByDate')
    html = _html(text)

    ul0, ul1, ul2 = html.xpath('//td/ul')

    flow = {
        int(k): v for k, v in [i.text.split(' : ') for i in ul0.xpath('.//li')]
    }
    yval = {k: v for k, v in [i.text.split(': ') for i in ul1.xpath('.//li')]}

    return {
        'input': {
            'Flow': flow,
        },
        'output': {
            'Flow': flow,
            'YVal': yval,
        },
    }


async def instrument_state() -> dict:
    """https://tsetmc.com/StaticContent/WS-InstrumentsState"""
    text = await _static_content('InstrumentsState')
    html = _html(text)

    tds = html.xpath('//td')
    flow = {
        int(k): v
        for k, v in [
            i.text_content().strip().split(' : ')
            for i in tds[7].xpath('.//li')
        ]
    }
    cetaval = {
        k.strip('"'): v.strip('"')
        for k, v in [
            i.text_content().strip().split(':') for i in tds[-1].xpath('.//p')
        ]
    }

    return {
        'input': {
            'Flow': flow,
        },
        'output': {
            'CEtaVal': cetaval,
        },
    }


async def instrument() -> dict:
    """https://tsetmc.com/StaticContent/WS-Instrument"""
    text = await _static_content('Instrument')
    html = _html(text)

    t0 = html.xpath('//table[1]')[0]
    assert t0 is not None
    t1, t2, t3 = t0.xpath('.//table')

    flow = {k: v for k, v in [i.text.split(' : ') for i in t1.xpath('.//li')]}

    yval_tds = [tr.xpath('.//td') for tr in t3.xpath('.//tr[not(.//hr)]')]
    yval = {
        k.text: (v1.text.strip(), v2.text.strip()) for k, v1, v2 in yval_tds
    }

    return {
        'input': {
            'Flow': flow,
        },
        'output': {
            'YVal': yval,
        },
    }


async def best_limits_all_ins() -> dict:
    """https://tsetmc.com/StaticContent/WS-BestLimitsAllIns"""
    text = await _static_content('BestLimitsAllIns')
    html = _html(text)

    t0 = html.xpath('//table[1]')[0]
    assert t0 is not None
    t1, t2 = t0.xpath('.//table')

    flow = {
        int(k): v
        for k, v in [i.text_content().split(' : ') for i in t1.xpath('.//li')]
    }
    out = {
        k.text_content().strip(): v.text_content().strip()
        for k, v in [tr.xpath('.//td') for tr in t2.xpath('.//tr')]
    }

    return {
        'input': {
            'Flow': flow,
        },
        'output': out,
    }


async def trade_last_day() -> dict:
    """https://tsetmc.com/StaticContent/WS-TradeLastDay"""
    text = await _static_content('TradeLastDay')
    html = _html(text)

    flow = {
        int(k): v
        for k, v in [
            i.text_content().split(' : ')
            for i in html.xpath('//ul[1]')[0].xpath('.//li')
        ]
    }
    out = {
        k.text_content().strip(): v.text_content().strip()
        for k, v in [
            tr.xpath('.//td')
            for tr in html.xpath('.//table[not(following-sibling::*)]//tr')
        ]
    }

    return {
        'input': {
            'Flow': flow,
        },
        'output': out,
    }



from asyncio import gather as _gather

from pandas import DataFrame as _Df, concat as _concat

from tsetmc import _logger
from tsetmc.instruments import Instrument as _Instrument, _LazyDS as LazyDS
from tsetmc.market_watch import market_watch_init as _market_watch_init

YVAL_EXCLUSIONS = {
    '306',  # مهرایران
    '327',  # شیشه01ن
}


def _dump(df: _Df):
    if not (l18 := df['l18']).is_unique:
        # happens when a symbol is promoted/demoted to another flow
        duplicated = l18.duplicated('last')
        df = df[~duplicated]
        _logger.info('Removed duplicated l18 values:\n%s', l18[duplicated])

    if not (codes := df.index).is_unique:
        duplicated = codes.duplicated('last')
        df = df[~duplicated]
        _logger.error('Removed duplicated codes:\n%s', codes[duplicated])

    df.sort_values('l18', inplace=True)
    df.to_csv(LazyDS.path, encoding='utf-8-sig', lineterminator='\n')


async def add_instrument(inst: _Instrument) -> None:
    # usually used in conjunction with Instrument.from_search
    code = inst.code
    info = await inst.info()
    df = LazyDS.df
    l18, l30 = info['lVal18AFC'], info['lVal30']
    new_row = [await inst.isin, await inst.cisin, l18, l30]
    try:
        row = df.loc[code]
    except KeyError:
        pass
    else:
        if [*row] == new_row:
            _logger.warning(f'{l18 = } already exists in dataset')
            return
    df.loc[code] = new_row
    _dump(df)


async def update(df: _Df | None = None) -> None:
    if df is None:
        mwi = await _market_watch_init(market_state=False, best_limits=False)
        df = mwi['prices']
    df = df[['l18', 'l30']][
        ~(
            df['yval'].isin(YVAL_EXCLUSIONS)
            | df['l18'].str.slice(-1).str.isdigit()
        )
    ]
    ds = LazyDS.df
    ds.update(df)  # update existing l18s/l30s
    new_items = df[~df.index.isin(ds.index)]

    if new_items.empty:
        _logger.info('No new entries were added by market watch.')
        return

    # add isin and cisin to new_items
    new_insts = [_Instrument(code) for code in new_items.index]
    await _gather(*[i.info() for i in new_insts])
    new_items['isin'] = [i._isin for i in new_insts]
    new_items['cisin'] = [i._cisin for i in new_insts]

    merged = LazyDS.df = _concat([ds, new_items])[
        ['isin', 'cisin', 'l18', 'l30']  # fix column order
    ]
    _dump(merged)
    _logger.info(f'{len(new_items)} new entries were added by market watch.')
